/*!
 * \file timer.cpp
 * \brief This file is used to take interval snapshot.
 * Copyright (C) 2011-2013 Nippon Telegraph and Telephone Corporation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */

#include <jni.h>
#include <jvmti.h>
#include <sys/time.h>

#include "util.hpp"
#include "agentThread.hpp"
#include "timer.hpp"

/*!
 * \brief TTimer constructor.
 * \param event     [in] Callback is used by interval calling.
 * \param interval  [in] Interval of timer.
 * \param timerName [in] Unique name of timer.
 */
TTimer::TTimer(TTimerEventFunc event, jlong interval, const char* timerName)
                                                     : TAgentThread(timerName) {
  
  /* Sanity check. */
  if (event == NULL) {
    throw "Event callback is NULL.";
  }
  
  /* Setting param. */
  this->timerInterval  = interval;
  this->_eventFunc     = event;
  this->_isInterrupted = false;
  
  /* Create semphore. */
  if (unlikely(sem_init(&this->notifySem, 0, 0))) {
    throw "Failure make semphore.";
  }
}

/*!
 * \brief TTimer destructor.
 */
TTimer::~TTimer(void) {
  
  /* Destroy semphore. */
  sem_destroy(&this->notifySem);
}

/*!
 * \brief JThread entry point called by interval or nofity.
 * \param jvmti [in] JVMTI environment object.
 * \param jni   [in] JNI environment object.
 * \param data  [in] Pointer of TTimer.
 */
void JNICALL TTimer::entryPointByInterval(jvmtiEnv *jvmti,
                                          JNIEnv *jni, void *data) {
  
  /* Get self. */
  TTimer *controller = (TTimer *)data;
  
  /* Create increment time. */
  struct timespec incTs = {0};
  incTs.tv_sec  = controller->timerInterval / 1000;
  incTs.tv_nsec = (controller->timerInterval % 1000) * 1000;
  
  /* Change running state. */
  controller->_isRunning = true;
  
  /* Loop for agent run. */
  while (!controller->_terminateRequest) {
    /* Reset timer interrupt flag. */
    controller->_isInterrupted = false;
    
    ENTER_PTHREAD_SECTION(&controller->mutex) {
      
      /* Create limit datetime. */
      struct timespec limitTs = {0};
      struct timeval  nowTv   = {0};
      gettimeofday(&nowTv, NULL);
      TIMEVAL_TO_TIMESPEC(&nowTv, &limitTs)
      limitTs.tv_sec  += incTs.tv_sec;
      limitTs.tv_nsec += incTs.tv_nsec;
      
      /* Wait for notification, termination or timeout. */
      pthread_cond_timedwait(&controller->mutexCond,
        &controller->mutex, &limitTs);
      
    } EXIT_PTHREAD_SECTION(&controller->mutex)
    
    /* If waiting finished by timeout. */
    if (!controller->_isInterrupted) {
      /* Call event callback. */
      (*controller->_eventFunc)(jvmti, jni, Interval);
    }
  }
  
  /* Change running state */
  controller->_isRunning = false;
}

/*!
 * \brief JThread entry point called by notify only.
 * \param jvmti [in] JVMTI environment object.
 * \param jni   [in] JNI environment object.
 * \param data  [in] Pointer of TTimer.
 */
void JNICALL TTimer::entryPointByCall(jvmtiEnv *jvmti, JNIEnv *jni, void *data) {
  
  /* Get self. */
  TTimer *controller = (TTimer *)data;
  
  /* Change running state. */
  controller->_isRunning = true;
  
  /* Loop for agent run. */
  while (!controller->_terminateRequest) {
    /* Reset timer interrupt flag. */
    controller->_isInterrupted = false;
    
    /* Wait notify or terminate. */
    sem_wait(&controller->notifySem);
    
    /* If waiting finished by timeout. */
    if (!controller->_isInterrupted) {
      /* Call event callback. */
      (*controller->_eventFunc)(jvmti, jni, Interval);
    }
  }
  
  /* Change running state */
  controller->_isRunning = false;
}

/*!
 * \brief Make and begin Jthread.
 * \param jvmti [in] JVMTI environment object.
 * \param env   [in] JNI environment object.
 */
void TTimer::start(jvmtiEnv *jvmti, JNIEnv *env) {
  
  /* Set entry callback. */
  if (this->timerInterval == 0) {
    
    /* Notify process mode. Using semaphore. */
    TAgentThread::start(jvmti, env, TTimer::entryPointByCall,
                                        this, JVMTI_THREAD_MAX_PRIORITY);
  } else {
    
    /* Interval process mode. Using pthread monitor. */
    TAgentThread::start(jvmti, env, TTimer::entryPointByInterval,
                                        this, JVMTI_THREAD_MAX_PRIORITY);
  }
}

/*!
 * \brief Notify reset timing to this thread from other thread.
 */
void TTimer::notify(void) {
  
  /* Notify follow entry callback type. */
  if (this->timerInterval == 0) {
    
    /* Notify process to waiting thread. */
    sem_post(&this->notifySem);
  } else {
    
    /* Set interrupt flag and notify. */
    ENTER_PTHREAD_SECTION(&this->mutex) {
      
      this->_isInterrupted = true;
      pthread_cond_signal(&this->mutexCond);
      
    } EXIT_PTHREAD_SECTION(&this->mutex)
  }
}

/*!
 * \brief Notify stopping to this thread from other thread.
 */
void TTimer::stop(void) {
  
  /* Terminate follow entry callback type. */
  if (this->timerInterval == 0) {
    
    /* Sanity check. */
    if (!this->_isRunning) {
      PRINT_WARN_MSG("AgentThread already finished.");
      return;
    }
    
    /* Send notification and count notify. */
    this->_isInterrupted = true;
    this->_terminateRequest = true;
    sem_post(&this->notifySem);
    
    /* SpinLock for AgentThread termination. */
    while (this->_isRunning) {
      ; /* none. */
    }
    
    /* Clean termination flag. */
    this->_terminateRequest = false;
  } else {
    
    /* Set interrupt flag and notify termination. */
    this->_isInterrupted = true;
    TAgentThread::stop();
  }
}

